import {
  Box,
  Text,
  Button,
  Grid,
  GridItem,
  Image,
  Input,
  Menu,
  MenuButton,
  MenuList,
  Radio,
  Stack,
  useDisclosure,
  useRadio,
  useRadioGroup,
  UseRadioGroupReturn,
  Flex,
  useOutsideClick,
  Portal,
} from "@chakra-ui/react";
import { ChevronDown } from "@feedloop/icon";
import React from "react";
import {
  defineComponent,
  useSourceHooks,
  useEvaluate,
  ListItemVariables,
  RuntimeContext,
  useRuntimeInput,
  useDisableCondition,
} from "@qorebase/app-component";
import { fallbackImg } from "@qorebase/app-runtime/src/utils/fallbackImage";
import InputComponentLabel from "@qorebase/app-runtime/src/components/InputComponentLabel";
import {
  InsightSchema,
  Row,
  TableSchema,
  ViewSchema,
} from "@qorebase/app-runtime/src/interfaces";
import { FixedSizeList } from "react-window";
import AutoSizer from "react-virtualized-auto-sizer";
import InfiniteLoader from "react-window-infinite-loader";
import { usePageParam } from "@qorebase/app-runtime/src/hooks/usePageParam";

const Option = (
  props: {
    item: Row;
    dataValue: string;
    onChange: (value: string) => void;
    setSelected: React.Dispatch<React.SetStateAction<string>>;
    onClose: () => void;
    search?: string;
    style: React.CSSProperties;
  } & typeof DropdownListComponent["Component"]["defaultProps"]
) => {
  const value = useEvaluate(
    props.properties.itemValue ||
      String(props.item.id) ||
      String(props.source.itemId)
  );
  const label = useEvaluate(props.properties.itemLabel);
  if (String(props.dataValue) === String(value)) props.setSelected(label);

  return (
    <Flex
      style={{ ...props.style }}
      onClick={() => {
        props.onClose();
        props.onChange(value);
      }}
      _hover={{
        bg: "gray.300",
      }}
      backgroundColor={String(props.dataValue) === String(value) && "gray.100"}
      paddingX="0.8rem"
      cursor={"pointer"}
      alignItems={"center"}
    >
      {label}
    </Flex>
  );
};

const RadioItem = React.memo(function RadioItem(
  props: {
    item: Row;
    dataValue: string;
    getRadioProps: UseRadioGroupReturn["getRadioProps"];
    isDisabled: boolean;
  } & typeof DropdownListComponent["Component"]["defaultProps"]
) {
  const value = useEvaluate(
    props.properties.itemValue ||
      String(props.item.id) ||
      String(props.source.itemId)
  );
  const { getInputProps, htmlProps, state, getCheckboxProps, getLabelProps } =
    useRadio(props.getRadioProps({ value }));
  const image = useEvaluate(props.properties.itemImage);
  const label = useEvaluate(props.properties.itemLabel);

  return (
    <GridItem
      bg="surface.900"
      rounded="md"
      border="1px"
      borderColor={
        state.isChecked ? props.app.theme.options.primaryColor : "gray.200"
      }
      style={{
        border: "1px solid rgb(226, 232, 240)",
        pointerEvents: props.isDisabled ? "none" : "auto",
        background: props.isDisabled && "#E7E7E7",
      }}
      _hover={
        !state.isChecked && {
          bg: "gray.200",
          borderColor: "gray.300",
        }
      }
    >
      {props.properties.style === "text_choice" && (
        <Radio
          w="full"
          p="1em"
          color={props.properties.style === "text_choice" && "gray.300"}
          colorScheme={
            props.properties.style === "text_choice"
              ? `var(--chakra-colors-qore-primary);`
              : undefined
          }
          {...props.getRadioProps({ value })}
        >
          <Text color={props.isDisabled && "#A2A2A2"}>{label}</Text>
        </Radio>
      )}
      {props.properties.style === "image_choice" && (
        <Box as="label" cursor="pointer" {...htmlProps}>
          <input {...getInputProps({})} hidden />
          <Stack
            direction={
              props.properties.layout === "horizontal_component"
                ? "row"
                : "column"
            }
            alignItems="center"
            justifyContent={
              props.properties.layout === "vertical_component" && "center"
            }
            spacing="1em"
            w="full"
            p="1em"
            _hover={
              !state.isChecked &&
              state.isHovered && { border: "1px solid gray.300" }
            }
            border="1px solid transparent"
            borderColor={
              state.isChecked && `var(--chakra-colors-qore-primary);`
            }
            borderRadius="md"
            {...getCheckboxProps()}
          >
            {props.properties.style === "image_choice" && (
              <Image
                src={image}
                alt={label}
                fallbackSrc={fallbackImg}
                w="64px"
                h="64px"
                borderRadius="md"
                style={{
                  border: "1px solid rgb(226, 232, 240)",
                  filter: props.isDisabled ? "grayscale(1)" : "grayscale(0)",
                  WebkitFilter: props.isDisabled
                    ? "grayscale(1)"
                    : "grayscale(0)",
                  opacity: props.isDisabled ? "0.5" : "1",
                }}
                {...getLabelProps()}
              />
            )}
            <Text
              display={props.properties.hideLabel && "none"}
              color={props.isDisabled && "#A2A2A2"}
            >
              {label}
            </Text>
          </Stack>
        </Box>
      )}
    </GridItem>
  );
});

const CustomRadioGroup = React.memo(function CustomRadioGroup(
  props: {
    isVerticalAlign: boolean;
    listData: Record<string, any>[];
    variableData: {
      onChange: (value: string) => void;
      value: string;
      validation: {
        error: string;
      };
    };
    isDisabled: boolean;
  } & typeof DropdownListComponent["Component"]["defaultProps"]
) {
  const { getRootProps, getRadioProps } = useRadioGroup({
    value: props.variableData.value,
    onChange: props.variableData.onChange,
  });

  return (
    <Grid
      templateColumns={
        props.isVerticalAlign
          ? "repeat(1,1fr)"
          : {
              base: "repeat(1,1fr)",
              md: "repeat(2,1fr)",
              lg: "repeat(3,1fr)",
            }
      }
      gap={2}
      {...getRootProps()}
    >
      {props.listData?.map((item) => (
        <ListItemVariables key={item.id} variables={item}>
          <RadioItem
            item={item}
            dataValue={props.variableData.value}
            getRadioProps={getRadioProps}
            isDisabled={props.isDisabled}
            {...props}
          />
        </ListItemVariables>
      ))}
    </Grid>
  );
});

const DropdownListComponent = defineComponent({
  name: "Dropdown List",
  type: "list",
  icon: "SelectInput",
  group: "input",
  display: "block",
  defaultProps: {
    label: "Dropdown List",
    placeholder: "Placeholder here",
    data: {
      inputID: "DropdownList",
      field: { type: "relation", options: {} },
      validation: {
        required: false,
      },
    },
    style: "dropdown",
    itemLabel: "",
    itemImage: "",
    itemValue: "",
    verticalAlignment: false,
    hideLabel: false,
    layout: "vertical_component",
  },
  propDefinition: {
    label: { group: "Design", type: "expression", options: {} },
    placeholder: {
      group: "Design",
      type: "expression",
      options: {},
    },
    data: {
      group: "Design",
      type: "state",
      options: {
        inputID: "DropdownList",
        field: { type: "relation", options: {} },
      },
    },
    style: {
      group: "Design",
      type: "string",
      options: {
        format: "select",
        options: [
          {
            label: "Dropdown",
            value: "dropdown",
          },
          {
            label: "Text Choice",
            value: "text_choice",
          },
          {
            label: "Image Choice",
            value: "image_choice",
          },
        ],
      },
    },
    itemLabel: {
      group: "Design",
      type: "expression",
      options: {},
      section: "dropdown",
    },
    itemImage: {
      group: "Design",
      type: "string",
      options: {
        format: "image",
      },
      section: "dropdown",
      hideOrShownOn: {
        type: "shownOn",
        expression: {
          type: "formula",
          formula: `style == "image_choice"`,
        },
      },
    },
    itemValue: {
      group: "Design",
      type: "expression",
      options: {},
      section: "dropdown",
    },
    verticalAlignment: {
      group: "Design",
      type: "boolean",
      options: {},
      section: "dropdown",
      hideOrShownOn: {
        type: "hideOn",
        expression: {
          type: "formula",
          formula: `style == "dropdown"`,
        },
      },
    },
    hideLabel: {
      group: "Design",
      type: "boolean",
      options: {},
      section: "dropdown",
      hideOrShownOn: {
        type: "shownOn",
        expression: {
          type: "formula",
          formula: `style == "image_choice"`,
        },
      },
    },
    layout: {
      group: "Design",
      section: "dropdown",
      hideOrShownOn: {
        type: "shownOn",
        expression: {
          type: "formula",
          formula: `style == "image_choice"`,
        },
      },
      type: "string",
      options: {
        format: "button",
        options: [
          {
            icon: "IconTopLeft",
            value: "vertical_component",
          },
          {
            icon: "IconLeft",
            value: "horizontal_component",
          },
        ],
      },
    },
  },
  Editor: async () => {
    return {
      data: (props) => {
        const pageParam = usePageParam();
        const currentPage = RuntimeContext.useSelectState(
          (state) => state.maps.page[pageParam]
        );
        const schema = RuntimeContext.useSelectState((state) => state.schema);
        const tablesAndViews = React.useMemo(
          (): (TableSchema | ViewSchema | InsightSchema)[] =>
            [].concat(schema.tables, schema.views, schema.insights),
          [schema]
        );
        const pageView = React.useMemo(
          () =>
            tablesAndViews.find((view) => view.name === currentPage?.sourceID),
          [currentPage?.sourceID]
        );
        console.log(
          "page",
          pageParam,
          currentPage,
          schema,
          tablesAndViews,
          pageView
        );
        return (
          <props.PropertyEditor.state
            propDef={props.propDef}
            propName={props.propName}
            value={props.value}
            onChange={props.onChange}
            view={pageView}
          />
        );
      },
    };
  },
  Component: (props) => {
    const { isOpen, onClose, onToggle } = useDisclosure();
    const boxWidth = React.useRef<null | HTMLButtonElement>(null);
    const [selected, setSelected] = React.useState<string>("");
    const [search, setSearch] = React.useState("");
    const [pageLoader, setPageLoader] = React.useState<number>(0);
    const sourceHooks = useSourceHooks({
      ...props.source,
      offset: 50 * pageLoader,
    });
    const [currentData, setCurrentData] = React.useState<Record<string, any>[]>(
      []
    );
    const { data: listData = [], status } = sourceHooks.useFetchList();

    const title = useEvaluate(props.properties.label);
    const placeholder = useEvaluate(props.properties.placeholder);
    const wrapperRef = React.useRef(null);
    const [searching, setSearching] = React.useState(false);
    useOutsideClick({
      enabled: searching ? false : true,
      ref: wrapperRef,
      handler: () => {
        onClose();
      },
    });

    const data = useRuntimeInput<"string">(
      props.properties.data,
      props.data.page.row,
      {
        validate: (value) => {
          if (typeof value === "undefined") {
            return { error: "" };
          } else if (!value) {
            return { error: "Required field, Please enter the data" };
          }
        },
      }
    );

    const validateData = (
      currentData: Record<string, any>[],
      newData: Record<string, any>[]
    ) => {
      if (!currentData.length) {
        return newData;
      } else {
        const filteredData: Record<string, any>[] = [...currentData];
        newData.forEach((newItem) => {
          const exist = currentData.find((item) => item.id === newItem.id);
          if (!exist) {
            filteredData.push(newItem);
          }
        });
        return filteredData;
      }
    };

    const getData = React.useCallback(() => {
      return validateData(currentData, listData);
    }, [currentData, listData]);

    const isVerticalAlign = React.useMemo(
      () => props.properties.verticalAlignment,
      [props.properties.verticalAlignment]
    );
    const styleType = React.useMemo(
      () => props.properties.style,
      [props.properties.style]
    );

    React.useEffect(() => {
      if (!!props.source.target && status === "success" && !!listData.length) {
        if (pageLoader === 0) {
          setCurrentData(listData);
        } else {
          setCurrentData(getData());
        }
      }
    }, [pageLoader, status, props.source.target, listData]);

    React.useEffect(() => {
      setCurrentData(listData);
      setPageLoader(0);
    }, [props.source.target, props.source.params]);

    // @ts-ignore
    const { isDisabled } = useDisableCondition(props.instance);
    const disabledProperties: React.CSSProperties = {
      border: "1px solid rgb(226, 232, 240)",
      pointerEvents: isDisabled ? "none" : "auto",
      background: isDisabled && "#E7E7E7",
      color: isDisabled && "#A2A2A2",
    };

    const identifier = React.useMemo(() => {
      if (search !== "") {
        return `${props.properties.itemLabel}`.match(/[^\b(item)[.]\w+/g)?.[0];
      }
      return null;
    }, [search]);

    const filteredData = React.useMemo(() => {
      return currentData.filter((item) => {
        if (search === "") {
          return item;
        } else if (
          item[identifier]?.toLowerCase().includes(search.toLocaleLowerCase())
        ) {
          return item;
        }
      });
    }, [currentData, search, identifier]);

    const isItemLoaded = React.useCallback(
      (index: number) => {
        if (index === currentData.length - 1 && listData.length) {
          return false;
        }
        return true;
      },
      [currentData, listData]
    );
    const loadMoreItems = React.useCallback(() => {
      return new Promise<void>((resolve) => {
        setPageLoader((prev) => prev + 1);
        resolve();
      });
    }, []);

    const handleLoadMore = React.useCallback(() => {
      if (status === "success") {
        setPageLoader((prev) => prev + 1);
      }
    }, [status]);

    const renderRow = React.useCallback(
      ({ index, style }) => {
        if (isItemLoaded(index) && listData) {
          return (
            <ListItemVariables
              key={filteredData[index].id}
              variables={filteredData[index]}
            >
              <Option
                item={filteredData[index]}
                onChange={data.onChange}
                dataValue={data.value}
                setSelected={setSelected}
                {...props}
                data-test="dropdown-list-option"
                search={search}
                style={style}
                onClose={onClose}
              />
            </ListItemVariables>
          );
        } else {
          return (
            <Flex
              paddingX="0.8rem"
              alignItems={"center"}
              height={"34px"}
              style={style}
            >
              Loading...
            </Flex>
          );
        }
      },
      [filteredData, data, setSelected]
    );

    const Options = React.useCallback(() => {
      return (
        <AutoSizer>
          {({ width, height }) => (
            <InfiniteLoader
              isItemLoaded={isItemLoaded}
              itemCount={filteredData.length}
              loadMoreItems={loadMoreItems}
            >
              {({ onItemsRendered, ref }) => (
                <FixedSizeList
                  height={height}
                  itemCount={filteredData.length}
                  itemSize={34}
                  width={width}
                  onItemsRendered={onItemsRendered}
                  ref={ref}
                >
                  {renderRow}
                </FixedSizeList>
              )}
            </InfiniteLoader>
          )}
        </AutoSizer>
      );
    }, [
      filteredData.length,
      isItemLoaded,
      loadMoreItems,
      props.source.params,
      props.source.target,
    ]);

    return (
      <Box>
        <InputComponentLabel
          title={title}
          name="dropdown-list"
          isRequired={props.properties.data.validation?.required}
          isEmpty={title.length === 0}
        />
        {styleType === "dropdown" && (
          <Box ref={wrapperRef}>
            <Menu autoSelect={false} isOpen={isOpen}>
              <MenuButton
                as={Button}
                style={disabledProperties}
                bgColor={"surface.900"}
                ref={boxWidth}
                textAlign="left"
                borderRadius="md"
                variant="outline"
                rightIcon={<ChevronDown />}
                width="100%"
                _focus={{ shadow: "0px" }}
                fontWeight="normal"
                data-test="dropdown-list-menu-button"
                onClick={() => {
                  onToggle();
                }}
              >
                {data.value ? selected : placeholder}
              </MenuButton>
              <Portal>
                <MenuList
                  border="0px"
                  width={boxWidth?.current?.offsetWidth}
                  zIndex={3}
                >
                  <Box paddingX="0.8rem">
                    <Input
                      value={search}
                      onChange={(e) => setSearch(e.target.value)}
                      placeholder="Search..."
                      mb={2}
                      onFocus={() => setSearching(true)}
                      onBlur={() => setSearching(false)}
                      onTouchStart={() => setSearching(true)}
                    />
                  </Box>
                  <Box maxHeight="11.5rem" overflowY="auto">
                    <Text
                      color="gray"
                      py={2}
                      px={4}
                      style={{ cursor: "default" }}
                    >
                      {placeholder}
                    </Text>
                    <div
                      style={{
                        width: "100%",
                        height: "100px",
                        flex: "1 1 auto",
                      }}
                    >
                      <Options />
                    </div>
                  </Box>
                </MenuList>
              </Portal>
            </Menu>
          </Box>
        )}
        {(styleType === "text_choice" || styleType === "image_choice") && (
          <Box>
            <CustomRadioGroup
              variableData={data}
              listData={currentData}
              isVerticalAlign={isVerticalAlign}
              isDisabled={isDisabled}
              {...props}
            />
            <Flex
              mt={2}
              justifyContent="center"
              w="full"
              flexDir={{ base: "column", lg: "row" }}
              display={
                listData.length && listData.length === 50 ? "flex" : "none"
              }
            >
              <Button
                variant="outline"
                bg="surface.900"
                onClick={handleLoadMore}
              >
                Load more
              </Button>
            </Flex>
          </Box>
        )}
        {!!data.validation.error && (
          <Text fontSize="sm" color="red" data-test="text-input-error">
            {data.validation.error}
          </Text>
        )}
      </Box>
    );
  },
});

export default DropdownListComponent;
